#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fnmatch.h>
#include <archive.h>
#include <archive_entry.h>

/*
 * Version: 20050509
 *
 * Example:
 *    cat base.?? | tarsplit freebsd-5.4.ctrl
 *
 * (C) Brian Candler 2005. Released under the BSD licence.
 */

static const char *o_prefix=0;
static const char *o_suffix=0;		/* suffix chosen by user */
static const char *o_suffix1=".tar";	/* default format suffix */
static const char *o_suffix2="";	/* default compression suffix */

static struct format_info {
	const char *name;
	int (*fn)(struct archive *);
	const char *def_suffix;
} format_opt[] = {
	{ "ustar",	archive_write_set_format_ustar,		".tar" },
	{ "cpio",	archive_write_set_format_cpio,		".cpio" },
	{ "pax",	archive_write_set_format_pax,		".pax" },
	{ 0,		0 }
};

static int set_compression_none(struct archive *ap)
{
	return ARCHIVE_OK;
}

static struct format_info compress_opt[] = {
	{ "none",	set_compression_none,			"" },
	{ "gzip",	archive_write_set_compression_gzip,	".gz" },
	{ "bzip2",	archive_write_set_compression_bzip2,	".bz2" },
	{ 0,		0 }
};

static int (*set_format)(struct archive *ap) = archive_write_set_format_ustar;
static int (*set_compression)(struct archive *ap) = set_compression_none;

static void choose_format(const char *o_format)
{
	struct format_info *f = &format_opt[0];

	set_format = 0;	
	while (f->name) {
		if (!strcmp(f->name, o_format)) {
			set_format = f->fn;
			o_suffix1 = f->def_suffix;
			break;
		}
		f++;
	}
	if (!set_format) {
		fprintf(stderr, "Unknown archive format '%s'\n", o_format);
		exit(1);
	}
}

static void choose_compression(const char *o_compress)
{
	struct format_info *c = &compress_opt[0];

	set_compression = 0;
	while (c->name) {
		if (!strcmp(c->name, o_compress)) {
			set_compression = c->fn;
			o_suffix2 = c->def_suffix;
			break;
		}
		c++;
	}
	if (!set_compression) {
		fprintf(stderr, "Unknown compression type '%s'\n", o_compress);
		exit(1);
	}
}

/*
 * Output archives
 */

struct outfile {
	const char *name;
	struct archive *archive;
};

/*
 * For each filename, record which file it should go into
 */

struct pattern {
	const char *pattern;
	int outfile;
	char used;
};

static int n_outfiles = 0;  /* actually used */
static int m_outfiles = 0;  /* memory allocated */
static struct outfile *outfiles = 0;

/* Exact patterns are kept in sorted order for fast bsearch */
static int n_patterns = 0;
static int m_patterns = 0;
static struct pattern *patterns = 0;

/* Wildcard patterns are searched linearly */
static int n_wild_patterns = 0;
static int m_wild_patterns = 0;
static struct pattern *wild_patterns = 0;

static int cmp_pattern(const void *p1, const void *p2)
{
	return strcmp(((const struct pattern *)p1)->pattern,
			((const struct pattern *)p2)->pattern);
}

static int find_outfile_by_name(const char *name)
{	
	int i;
	struct archive *ap;
	char fnbuf[MAXPATHLEN];

	for (i=0; i<n_outfiles; i++)
		if (!strcmp(name, outfiles[i].name)) return i;

	if (n_outfiles >= m_outfiles) {
		m_outfiles += 10;
		outfiles = realloc(outfiles, m_outfiles * sizeof (struct outfile));
	}

	i = n_outfiles++;
	outfiles[i].name = strdup(name);
	if (name[0] == '\0') {
		outfiles[i].archive = 0;
		return i;
	}

	ap = archive_write_new();
	outfiles[i].archive = ap;
	set_format(ap);
	set_compression(ap);
	
	snprintf(fnbuf, sizeof(fnbuf), "%s%s%s%s", o_prefix?o_prefix:"",
		name,
		o_suffix ? o_suffix : o_suffix1,
		o_suffix ? "" : o_suffix2);

	if (archive_write_open_file(ap, fnbuf) != ARCHIVE_OK) {
		fprintf(stderr, "%s: %s\n", fnbuf, archive_error_string(ap));
		exit(1);
	}
	return i;
}

static int alloc_pattern(void)
{
	if (n_patterns >= m_patterns) {
		m_patterns += 100;
		patterns = realloc(patterns, m_patterns * sizeof (struct pattern));
	}
	return n_patterns++;
}

static int alloc_wild_pattern(void)
{
	if (n_wild_patterns >= m_wild_patterns) {
		m_wild_patterns += 10;
		wild_patterns = realloc(wild_patterns, m_wild_patterns * sizeof (struct pattern));
	}
	return n_wild_patterns++;
}

static int compar_pattern(const void *key, const void *pat)
{
	return strcmp((const char *)key, ((const struct pattern *)pat)->pattern);
}

static int find_outfile_from_pattern(const char *name)
{
	int i;
	struct pattern *p = bsearch(name, patterns, n_patterns,
		sizeof (struct pattern), compar_pattern);

	if (p) {
		p->used = 1;
		return p->outfile;
	}

	/* Try again with linear search through wildcards */
	for (i=0,p=wild_patterns; i<n_wild_patterns; i++,p++) {
		if (!fnmatch(p->pattern, name, 0)) {
			p->used = 1;
			return p->outfile;
		}
	}

	printf("#?\t%s\n", name);
	return -1;
}

static void run(const char *filename)
{
	struct archive *ap;
	struct archive_entry *entry;
	
	ap = archive_read_new();
	if (archive_read_support_compression_all(ap) ||
	    archive_read_support_format_all(ap) ||
	    archive_read_open_file(ap, filename, 10240)) {
		fprintf(stderr, "%s: %s\n", filename?filename:"(stdin)",
		archive_error_string(ap));
		exit(1);
	}
	while (archive_read_next_header(ap, &entry) == ARCHIVE_OK) {
		const char *pathname = archive_entry_pathname(entry);
		const char *hardlink = archive_entry_hardlink(entry);
		int of = find_outfile_from_pattern(pathname);
		struct archive *dest;
		off_t wpos = 0;

		if (hardlink) {
			/* try to ensure that all hardlinked files go
			   into the same output bin. Note: we assume that
			   all hardlinks point to the same original file,
			   i.e. we don't have 'a', 'b->a', 'c->b' etc */
			int of2 = find_outfile_from_pattern(hardlink);
			if (of2 != of) {
				of = of2;
				printf("###hardlink to %s\n", hardlink);
				printf("#%s\t%s\n",
					(of < 0) ? "" : outfiles[of].name,
					pathname);
			}
		}

		if (of < 0) continue;

		dest = outfiles[of].archive;
		if (!dest) {
			archive_read_data_skip(ap);
			continue;
		}

		archive_write_header(dest, entry);

		while(1) {
			const void *bufp;
			size_t len;
			off_t off;
			int res = archive_read_data_block(ap, &bufp, &len, &off);
			
			if (res == ARCHIVE_EOF) {
				break;
			}
			else if (res == ARCHIVE_OK) {
				while (off > wpos) {
					static char zero[1] = "";
					archive_write_data(dest, zero, 1);
					wpos++;
				}
				archive_write_data(dest, bufp, len);
				wpos += len;
			}
			else if (res == ARCHIVE_WARN) {
				fprintf(stderr, "Warning: %s\n", archive_error_string(ap));
			}
			else {
				fprintf(stderr, "Error: %s\n", archive_error_string(ap));
				break;
			}
		}
	}
	archive_read_finish(ap);
}

int main(int argc, const char *argv[])
{
	FILE *ff;
	char buf[MAXPATHLEN*2];
	int i;

	if (argc < 2) {
		fprintf(stderr, "Usage: tarsplit controlfile [inputarchive...]\n");
		return (1);
	}

	if (!strcmp(argv[1], "-"))
		ff = stdin;
	else
		ff = fopen(argv[1],"r");
	if (!ff) {
		perror(argv[1]);
		exit(1);
	}
	
	while (fgets(buf, sizeof(buf), ff)) {
		char *p = buf;
		int pat;
		int len;

		/* strip trailing \n and whitespace */
		len = strlen(buf);
		while (len > 0 && isspace(buf[len-1]))
			buf[--len] = '\0';

		if (buf[0] == '#') continue;

		while (*p && !isspace(*p)) p++;
		if (*p) {
			*p++ = '\0';
			while (isspace(*p)) p++;
		}
		if (*p == '\0') {
			if (buf[0])
				fprintf(stderr, "Incomplete line ignored: %s\n", buf);
			continue;
		}

		if (buf[0] == '*') {
			if (n_outfiles > 0) {
				fprintf(stderr, "Error: %s: options must come before filename patterns\n", buf);
				exit(1);
			}
			if (!strcmp(buf, "*PREFIX"))
				o_prefix = strdup(p);
			else if (!strcmp(buf, "*SUFFIX"))
				o_suffix = strdup(p);
			else if (!strcmp(buf, "*FORMAT"))
				choose_format(p);
			else if (!strcmp(buf, "*COMPRESS"))
				choose_compression(p);
			else
				fprintf(stderr, "Unknown option ignored: %s\n", buf);
			continue;
		}

		if (p[0] == '^') {
			pat = alloc_wild_pattern();
			wild_patterns[pat].pattern = strdup(p+1);
			wild_patterns[pat].outfile = find_outfile_by_name(buf);
			wild_patterns[pat].used = 0;
		} else {
			pat = alloc_pattern();
			patterns[pat].pattern = strdup(p);
			patterns[pat].outfile = find_outfile_by_name(buf);
			patterns[pat].used = 0;
		}
	}
	if (ferror(ff)) {
		perror(argv[1]);
		exit(1);
	}
	fclose(ff);

	qsort(patterns, n_patterns, sizeof (struct pattern), cmp_pattern);

	if (argc < 3)
		run(0);
	else {
		for (i=2; i<argc; i++)
			run(argv[i]);
	}
	
	/* close all outfiles */
	for (i=0; i<n_outfiles; i++)
		archive_write_finish(outfiles[i].archive);

	/* report on unused patterns */
	for (i=0; i<n_patterns; i++)
		if (!patterns[i].used)
			printf("#unused:%s\t%s\n",
				outfiles[patterns[i].outfile].name,
				patterns[i].pattern);

	for (i=0; i<n_wild_patterns; i++)
		if (!wild_patterns[i].used)
			printf("#unused:%s\t^%s\n",
				outfiles[wild_patterns[i].outfile].name,
				wild_patterns[i].pattern);

	return 0;
}
